@@ -1,7 +1,7 @@
 /* Trampoline construction */
 
 /*
- * Copyright 1995-2017 Bruno Haible <bruno@clisp.org>
+ * Copyright 1995-2022 Bruno Haible <bruno@clisp.org>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 
@@ -72,34 +72,29 @@ extern void (*tramp) (); /* trampoline prototype */
 #endif
 #endif
 
-#ifndef CODE_EXECUTABLE
+#ifndef CODE_EXECUTABLE                 /* Linux, Hurd, macOS, *BSD, IRIX, Solaris, Cygwin, native Windows */
   /* How do we make the trampoline's code executable? */
-  #if defined(HAVE_MACH_VM) || defined(HAVE_WORKING_MPROTECT)
-    #if HAVE_MPROTECT_AFTER_MALLOC_CAN_EXEC > 0
+  #if HAVE_WORKING_MPROTECT             /* Linux, Hurd, macOS, *BSD, IRIX, Solaris, Cygwin */
+    #if HAVE_MPROTECT_AFTER_MALLOC_CAN_EXEC > 0    /* Hurd, macOS, *BSD except HardenedBSD, IRIX, Solaris, Cygwin */
       /* mprotect() [or equivalent] the malloc'ed area. */
       #define EXECUTABLE_VIA_MALLOC_THEN_MPROTECT
     #elif HAVE_MPROTECT_AFTER_MMAP_CAN_EXEC > 0
       /* mprotect() [or equivalent] the mmap'ed area. */
       #define EXECUTABLE_VIA_MMAP_THEN_MPROTECT
-    #elif defined(HAVE_MMAP_SHARED_CAN_EXEC)
+    #elif HAVE_MMAP_SHARED_CAN_EXEC                /* Linux, HardenedBSD */
       #define EXECUTABLE_VIA_MMAP_FILE_SHARED
     #else
       #error "Don't know how to make memory pages executable."
     #endif
-  #else
-    #ifdef HAVE_MMAP
+  #else                                            /* native Windows, very old Linux */
+    #if HAVE_MMAP_ANONYMOUS                        /* very old Linux */
       /* Use an mmap'ed page. */
       #define EXECUTABLE_VIA_MMAP
     #else
-      #ifdef HAVE_SHM
-        /* Use an shmat'ed page. */
-        #define EXECUTABLE_VIA_SHM
+      #if defined(_WIN32) && ! defined(__CYGWIN__) /* native Windows */
+        #define EXECUTABLE_VIA_VIRTUALALLOC
       #else
-        #if (defined(_WIN32) || defined(__WIN32__)) && ! defined(__CYGWIN__) /* native Windows */
-          #define EXECUTABLE_VIA_VIRTUALALLOC
-        #else
-          ??
-        #endif
+        ??
       #endif
     #endif
   #endif
@@ -110,70 +105,42 @@ extern void (*tramp) (); /* trampoline prototype */
 #include <sys/types.h>
 #include <stdlib.h> /* declares abort(), malloc(), free() */
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+#include <unistd.h> /* declares getpagesize() */
 #endif
 
 /* Define intptr_t, uintptr_t. */
 #include <stdint.h>
 
 /* Declare getpagesize(). */
-#ifdef HAVE_GETPAGESIZE
+/* On HP-UX, getpagesize exists, but it is not declared in <unistd.h> even if
+   the compiler options -D_HPUX_SOURCE -D_XOPEN_SOURCE=600 are used.  */
+#ifdef __hpux
+extern
 #ifdef __cplusplus
-extern "C" RETGETPAGESIZETYPE getpagesize (void);
-#else
-extern RETGETPAGESIZETYPE getpagesize (void);
+       "C"
 #endif
-#else
-#ifdef HAVE_SYS_PARAM_H
-#include <sys/param.h>
-#else
-/* Not Unix, e.g. mingw32 */
-#define PAGESIZE 4096
-#endif
-#define getpagesize() PAGESIZE
+       int getpagesize (void);
 #endif
 
-/* Declare mprotect() or equivalent. */
+/* Declare mprotect(). */
 #if defined(EXECUTABLE_VIA_MALLOC_THEN_MPROTECT) || defined(EXECUTABLE_VIA_MMAP_THEN_MPROTECT)
-#ifdef HAVE_MACH_VM
-#include <sys/resource.h>
-#include <mach/mach_interface.h>
-#ifdef __osf__
-#include <mach_init.h>
-#endif
-#include <mach/machine/vm_param.h>
-#else
 #include <sys/types.h>
 #include <sys/mman.h>
 #endif
-#endif
 
 /* Declare mmap(). */
 #if defined(EXECUTABLE_VIA_MMAP) || defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
 #include <sys/types.h>
 #include <sys/mman.h>
-#if !defined(PROT_EXEC) && defined(PROT_EXECUTE) /* Irix 4.0.5 needs this */
-#define PROT_EXEC PROT_EXECUTE
-#endif
 #endif
 
 /* Declare open(). */
-#if defined(EXECUTABLE_VIA_MMAP_DEVZERO) || defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
+#if defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
 #include <sys/types.h>
 #include <unistd.h>
 #include <fcntl.h>
 #endif
 
-/* Declare shmget(), shmat(), shmctl(). */
-#ifdef EXECUTABLE_VIA_SHM
-#include <sys/types.h>
-#include <sys/ipc.h>
-#include <sys/shm.h>
-#ifdef HAVE_SYS_SYSMACROS_H
-#include <sys/sysmacros.h>
-#endif
-#endif
-
 /* Declare VirtualAlloc(), GetSystemInfo. */
 #ifdef EXECUTABLE_VIA_VIRTUALALLOC
 #define WIN32_LEAN_AND_MEAN
@@ -181,11 +148,6 @@ extern RETGETPAGESIZETYPE getpagesize (void);
 #include <windows.h>
 #endif
 
-/* Some old mmap() implementations require the flag MAP_FILE whenever you pass
-   an fd >= 0. */
-#ifndef MAP_FILE
-#define MAP_FILE 0
-#endif
 /* Some old mmap() implementations require the flag MAP_VARIABLE whenever you
    pass an addr == NULL. */
 #ifndef MAP_VARIABLE
@@ -193,14 +155,25 @@ extern RETGETPAGESIZETYPE getpagesize (void);
 #endif
 
 /* Support for instruction cache flush. */
+#if defined _WIN32 && !defined __CYGWIN__
+# define WIN32_LEAN_AND_MEAN
+# define WIN32_EXTRA_LEAN
+# include <windows.h>
+#elif defined __APPLE__ && defined __MACH__
+# include <libkern/OSCacheControl.h>
+#elif defined _AIX
+# include <sys/cache.h>
+#elif defined __sgi
+# include <sys/cachectl.h>
+#else
 #ifdef __i386__
-#if (defined(_WIN32) || defined(__WIN32__)) && ! defined(__CYGWIN__) /* native Windows */
+#if defined(_WIN32) && ! defined(__CYGWIN__) /* native Windows */
 #define WIN32_LEAN_AND_MEAN
 #define WIN32_EXTRA_LEAN
 #include <windows.h>
 #endif
 #endif
-#if defined(__mips__) || defined(__mipsn32__) || defined(__mips64__)
+#if defined(__mips__) || defined(__mipsn32__) || defined(__mips64__) || defined(__riscv32__) || defined(__riscv64__)
 #ifdef HAVE_SYS_CACHECTL_H /* IRIX, Linux */
 #include <sys/cachectl.h>
 #else
@@ -217,10 +190,17 @@ extern void __TR_clear_cache_4();
 extern void __TR_clear_cache();
 #endif
 #endif
+#endif
 
 /* Support for multithread-safe coding. */
 #include "glthread/lock.h"
 
+#if defined(__APPLE__) && defined(__MACH__) /* only needed on macOS */
+#define KEEP_TEMP_FILE_VISIBLE
+/* Support for temporary files that are cleaned up automatically. */
+#include "clean-temp-simple.h"
+#endif
+
 #if !defined(CODE_EXECUTABLE) && defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
 /* Opens a file descriptor and attempts to make it non-inheritable. */
 static int open_noinherit (const char *filename, int flags, int mode)
@@ -342,6 +322,18 @@ static int open_noinherit (const char *filename, int flags, int mode)
 #define TRAMP_LENGTH 64
 #define TRAMP_ALIGN 8
 #endif
+#if defined(__riscv32__)
+#define TRAMP_LENGTH 36
+#define TRAMP_ALIGN 4
+#endif
+#if defined(__riscv64__)
+#define TRAMP_LENGTH 48
+#define TRAMP_ALIGN 8
+#endif
+#if defined(__loongarch64__)
+#define TRAMP_LENGTH 48
+#define TRAMP_ALIGN 8
+#endif
 
 #ifndef TRAMP_BIAS
 #define TRAMP_BIAS 0
@@ -351,38 +343,40 @@ static int open_noinherit (const char *filename, int flags, int mode)
 static long pagesize = 0;
 #endif
 
-#if !defined(CODE_EXECUTABLE) && (defined(EXECUTABLE_VIA_MMAP_DEVZERO) || defined(EXECUTABLE_VIA_MMAP_FILE_SHARED))
+#if !defined(CODE_EXECUTABLE) && defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
 
 /* Variables needed for obtaining memory pages via mmap(). */
-#if defined(EXECUTABLE_VIA_MMAP_DEVZERO)
-static int zero_fd;
-#endif
-#if defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
 static int file_fd;
 static long file_length;
-#endif
 
 /* Initialization of these variables. */
 static void for_mmap_init (void)
 {
-#if defined(EXECUTABLE_VIA_MMAP_DEVZERO)
-  zero_fd = open("/dev/zero",O_RDONLY,0644);
-  if (zero_fd < 0)
-    { fprintf(stderr,"trampoline: Cannot open /dev/zero!\n"); abort(); }
-#endif
-#if defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
   {
     char filename[100];
     sprintf(filename, "%s/trampdata-%d-%ld", "/tmp", getpid (), random ());
+#if defined(KEEP_TEMP_FILE_VISIBLE)
+    if (register_temporary_file(filename) < 0)
+      { fprintf(stderr,"trampoline: Out of virtual memory!\n"); abort(); }
+#endif
     file_fd = open_noinherit (filename, O_CREAT | O_RDWR | O_TRUNC, 0700);
     if (file_fd < 0)
-      { fprintf(stderr,"trampoline: Cannot open %s!\n",filename); abort(); }
+      {
+#if defined(KEEP_TEMP_FILE_VISIBLE)
+        unregister_temporary_file(filename);
+#endif
+        fprintf(stderr,"trampoline: Cannot open %s!\n",filename);
+        abort();
+      }
+#if !defined(KEEP_TEMP_FILE_VISIBLE)
     /* Remove the file from the file system as soon as possible, to make
-       sure there is no leftover after this process terminates or crashes. */
+       sure there is no leftover after this process terminates or crashes.
+       On macOS 11.2, this does not work: It would make the mmap call below,
+       with arguments PROT_READ|PROT_EXEC and MAP_SHARED, fail. */
     unlink(filename);
+#endif
   }
   file_length = 0;
-#endif
 }
 
 /* Once-only initializer for these variables. */
@@ -410,6 +404,11 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
   /* First, get the page size once and for all. */
   if (!pagesize)
     {
+#if defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
+      /* Use a once-only initializer here, since simultaneous execution of
+         for_mmap_init() in multiple threads must be avoided. */
+      gl_once (for_mmap_once, for_mmap_init);
+#endif
       /* Simultaneous execution of this initialization in multiple threads
          is OK. */
 #if defined(EXECUTABLE_VIA_VIRTUALALLOC)
@@ -419,15 +418,8 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
       SYSTEM_INFO info;
       GetSystemInfo(&info);
       pagesize = info.dwPageSize;
-#elif defined(HAVE_MACH_VM)
-      pagesize = vm_page_size;
 #else
       pagesize = getpagesize();
-#endif
-#if defined(EXECUTABLE_VIA_MMAP_DEVZERO) || defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
-      /* Use a once-only initializer here, since simultaneous execution of
-         for_mmap_init() in multiple threads must be avoided. */
-      gl_once (for_mmap_once, for_mmap_init);
 #endif
     }
 #endif
@@ -467,29 +459,12 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
       page = (char*)(((uintptr_t)page + sizeof(intptr_t) + TRAMP_ALIGN-1) & -TRAMP_ALIGN);
 #else
 #ifdef EXECUTABLE_VIA_MMAP_THEN_MPROTECT
-#ifdef HAVE_MMAP_ANONYMOUS
       /* Use mmap with the MAP_ANONYMOUS or MAP_ANON flag. */
       page = mmap(NULL, pagesize, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS | MAP_VARIABLE, -1, 0);
-#else
-      /* Use mmap on /dev/zero. */
-      page = mmap(NULL, pagesize, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_FILE | MAP_VARIABLE, zero_fd, 0);
-#endif
 #endif
 #ifdef EXECUTABLE_VIA_MMAP
-#ifdef HAVE_MMAP_ANONYMOUS
       /* Use mmap with the MAP_ANONYMOUS or MAP_ANON flag. */
       page = mmap(NULL, pagesize, PROT_READ | PROT_WRITE | PROT_EXEC, MAP_PRIVATE | MAP_ANONYMOUS | MAP_VARIABLE, -1, 0);
-#else
-      /* Use mmap on /dev/zero. */
-      page = mmap(NULL, pagesize, PROT_READ | PROT_WRITE | PROT_EXEC, MAP_PRIVATE | MAP_FILE | MAP_VARIABLE, zero_fd, 0);
-#endif
-#endif
-#ifdef EXECUTABLE_VIA_SHM
-      int shmid = shmget(IPC_PRIVATE, pagesize, 0700|IPC_CREAT);
-      if (shmid<0)
-        { page = (char*)(-1); }
-      else
-        { page = shmat(shmid, 0, 0); shmctl(shmid, IPC_RMID, 0); }
 #endif
       if (page == (char*)(-1))
         { fprintf(stderr,"trampoline: Out of virtual memory!\n"); abort(); }
@@ -577,7 +552,7 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
    *    				AC 22 lo16(<variable>)
    *    li $25,<address>&0xffff0000	3C 19 hi16(<address>)
    *    ori $25,$25,<address>&0xffff	37 39 lo16(<address>)
-   *    j $25				03 20 00 08
+   *    jal $0,$25			03 20 00 09  was:  j $25   03 20 00 08
    *    nop				00 00 00 00
    */
   /* What about big endian / little endian ?? */
@@ -593,7 +568,7 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
   *(short *) (function +18) = (unsigned long) address >> 16;
   *(short *) (function +20) = 0x3739;
   *(short *) (function +22) = (unsigned long) address & 0xffff;
-  *(long *)  (function +24) = 0x03200008;
+  *(long *)  (function +24) = 0x03200009;
   *(long *)  (function +28) = 0x00000000;
 #define is_tramp(function)  \
   *(unsigned short *) (function + 0) == 0x3C02 && \
@@ -602,7 +577,7 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
   *(unsigned short *) (function +12) == 0xAC22 && \
   *(unsigned short *) (function +16) == 0x3C19 && \
   *(unsigned short *) (function +20) == 0x3739 && \
-  *(unsigned long *)  (function +24) == 0x03200008 && \
+  *(unsigned long *)  (function +24) == 0x03200009 && \
   *(unsigned long *)  (function +28) == 0x00000000
 #define hilo(hiword,loword)  \
   (((unsigned long) (hiword) << 16) | (unsigned long) (loword))
@@ -619,7 +594,7 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
    *    lw $3,28($25)			8F 23 00 1C
    *    sw $3,0($2)			AC 43 00 00
    *    lw $25,32($25)			8F 39 00 20
-   *    j $25				03 20 00 08
+   *    jal $0,$25			03 20 00 09  was:  j $25   03 20 00 08
    *    nop				00 00 00 00
    *    .word <variable>		<variable>
    *    .word <data>			<data>
@@ -629,7 +604,7 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
   *(unsigned int *) (function + 4) = 0x8F23001C;
   *(unsigned int *) (function + 8) = 0xAC430000;
   *(unsigned int *) (function +12) = 0x8F390020;
-  *(unsigned int *) (function +16) = 0x03200008;
+  *(unsigned int *) (function +16) = 0x03200009;
   *(unsigned int *) (function +20) = 0x00000000;
   *(unsigned int *) (function +24) = (unsigned int) variable;
   *(unsigned int *) (function +28) = (unsigned int) data;
@@ -640,7 +615,7 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
   *(int *)          (function + 4) == 0x8F23001C && \
   *(int *)          (function + 8) == 0xAC430000 && \
   *(int *)          (function +12) == 0x8F390020 && \
-  *(int *)          (function +16) == 0x03200008 && \
+  *(int *)          (function +16) == 0x03200009 && \
   *(int *)          (function +20) == 0x00000000
 #define tramp_address(function)  \
   *(unsigned int *) (function +32)
@@ -670,7 +645,7 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
    *					37 39 hi16(lo32(<address>))
    *					00 19 CC 38
    *					37 39 lo16(lo32(<address>))
-   *    j $25				03 20 00 08
+   *    jal $0,$25			03 20 00 09  was:  j $25   03 20 00 08
    *    nop				00 00 00 00
    */
   /* What about big endian / little endian ?? */
@@ -705,7 +680,7 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
   *(int *)   (function +68) = 0x0019CC38;
   *(short *) (function +72) = 0x3739;
   *(short *) (function +74) = (unsigned long) address & 0xffff;
-  *(int *)   (function +76) = 0x03200008;
+  *(int *)   (function +76) = 0x03200009;
   *(int *)   (function +80) = 0x00000000;
 #define is_tramp(function)  \
   *(unsigned short *) (function + 0) == 0x3C02 && \
@@ -727,7 +702,7 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
   *(unsigned short *) (function +64) == 0x3739 && \
   *(unsigned int *)   (function +68) == 0x0019CC38 && \
   *(unsigned short *) (function +72) == 0x3739 && \
-  *(unsigned int *)   (function +76) == 0x03200008 && \
+  *(unsigned int *)   (function +76) == 0x03200009 && \
   *(unsigned int *)   (function +80) == 0x00000000
 #define hilo(word3,word2,word1,word0)  \
   (((unsigned long) (word3) << 48) | ((unsigned long) (word2) << 32) | \
@@ -754,23 +729,29 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
    *    ld $3,32($25)			DF 23 00 20
    *    sd $3,0($2)			FC 43 00 00
    *    ld $25,40($25)			DF 39 00 28
-   *    j $25				03 20 00 08
+   *    jal $0,$25			03 20 00 09  was:  j $25   03 20 00 08
    *    nop				00 00 00 00
    *    .dword <variable>		<variable>
    *    .dword <data>			<data>
    *    .dword <address>		<address>
    */
-  *(long *)          (function + 0) = 0xDF220018DF230020L;
-  *(long *)          (function + 8) = 0xFC430000DF390028L;
-  *(long *)          (function +16) = 0x0320000800000000L;
+  *(unsigned int *)  (function + 0) = 0xDF220018;
+  *(unsigned int *)  (function + 4) = 0xDF230020;
+  *(unsigned int *)  (function + 8) = 0xFC430000;
+  *(unsigned int *)  (function +12) = 0xDF390028;
+  *(unsigned int *)  (function +16) = 0x03200009;
+  *(unsigned int *)  (function +20) = 0x00000000;
   *(unsigned long *) (function +24) = (unsigned long) variable;
   *(unsigned long *) (function +32) = (unsigned long) data;
   *(unsigned long *) (function +40) = (unsigned long) address;
 #define TRAMP_CODE_LENGTH  24
 #define is_tramp(function)  \
-  *(long *)          (function + 0) == 0xDF220018DF230020L && \
-  *(long *)          (function + 8) == 0xFC430000DF390028L && \
-  *(long *)          (function +16) == 0x0320000800000000L
+  *(unsigned int *)  (function + 0) == 0xDF220018 && \
+  *(unsigned int *)  (function + 4) == 0xDF230020 && \
+  *(unsigned int *)  (function + 8) == 0xFC430000 && \
+  *(unsigned int *)  (function +12) == 0xDF390028 && \
+  *(unsigned int *)  (function +16) == 0x03200009 && \
+  *(unsigned int *)  (function +20) == 0x00000000
 #define tramp_address(function)  \
   *(unsigned long *) (function +40)
 #define tramp_variable(function)  \
@@ -1465,6 +1446,114 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
 #define tramp_data(function)  \
   (*(unsigned long *) (function +40))
 #endif
+#ifdef __riscv32__
+  /* function:
+   *    auipc t3,0			00000E17
+   *    lw t0,24(t3)			018E2283
+   *    lw t1,28(t3)			01CE2303
+   *    lw t2,32(t3)			020E2383
+   *    sw t0,(t1)			00532023
+   *    jr t2				00038067
+   * data:     .word <data>
+   * variable: .word <variable>
+   * address:  .word <address>
+   */
+  *(int *)   (function + 0) = 0x00000E17;
+  *(int *)   (function + 4) = 0x018E2283;
+  *(int *)   (function + 8) = 0x01CE2303;
+  *(int *)   (function +12) = 0x020E2383;
+  *(int *)   (function +16) = 0x00532023;
+  *(int *)   (function +20) = 0x00038067;
+  *(int *)   (function +24) = (unsigned int) data;
+  *(int *)   (function +28) = (unsigned int) variable;
+  *(int *)   (function +32) = (unsigned int) address;
+#define TRAMP_CODE_LENGTH  24
+#define is_tramp(function)  \
+  *(unsigned int *) (function + 0) == 0x00000E17 && \
+  *(unsigned int *) (function + 4) == 0x018E2283 && \
+  *(unsigned int *) (function + 8) == 0x01CE2303 && \
+  *(unsigned int *) (function +12) == 0x020E2383 && \
+  *(unsigned int *) (function +16) == 0x00532023 && \
+  *(unsigned int *) (function +20) == 0x00038067
+#define tramp_address(function)  \
+  (*(unsigned int *) (function +32))
+#define tramp_variable(function)  \
+  (*(unsigned int *) (function +28))
+#define tramp_data(function)  \
+  (*(unsigned int *) (function +24))
+#endif
+#ifdef __riscv64__
+  /* function:
+   *    auipc t3,0			00000E17
+   *    ld t0,24(t3)			018E3283
+   *    ld t1,32(t3)			020E3303
+   *    ld t2,40(t3)			028E3383
+   *    sd t0,(t1)			00533023
+   *    jr t2				00038067
+   * data:     .quad <data>
+   * variable: .quad <variable>
+   * address:  .quad <address>
+   */
+  *(int *)   (function + 0) = 0x00000E17;
+  *(int *)   (function + 4) = 0x018E3283;
+  *(int *)   (function + 8) = 0x020E3303;
+  *(int *)   (function +12) = 0x028E3383;
+  *(int *)   (function +16) = 0x00533023;
+  *(int *)   (function +20) = 0x00038067;
+  *(long *)  (function +24) = (unsigned long) data;
+  *(long *)  (function +32) = (unsigned long) variable;
+  *(long *)  (function +40) = (unsigned long) address;
+#define TRAMP_CODE_LENGTH  24
+#define is_tramp(function)  \
+  *(unsigned int *) (function + 0) == 0x00000E17 && \
+  *(unsigned int *) (function + 4) == 0x018E3283 && \
+  *(unsigned int *) (function + 8) == 0x020E3303 && \
+  *(unsigned int *) (function +12) == 0x028E3383 && \
+  *(unsigned int *) (function +16) == 0x00533023 && \
+  *(unsigned int *) (function +20) == 0x00038067
+#define tramp_address(function)  \
+  (*(unsigned long *) (function +40))
+#define tramp_variable(function)  \
+  (*(unsigned long *) (function +32))
+#define tramp_data(function)  \
+  (*(unsigned long *) (function +24))
+#endif
+#ifdef __loongarch64__
+  /* function:
+   *    pcaddu12i $r12,0		1C00000C
+   *    ld.d $r13,$r12,24		28C0618D
+   *    ld.d $r14,$r12,32		28C0818E
+   *    st.d $r14,$r13,0		29C001AE
+   *    ld.d $r12,$r12,40		28C0A18C
+   *    jirl $r0,$r12,0			4C000180
+   *    .dword <variable>		<variable>
+   *    .dword <data>			<data>
+   *    .dword <address>		<address>
+   */
+  *(unsigned int *)  (function + 0) = 0x1C00000C;
+  *(unsigned int *)  (function + 4) = 0x28C0618D;
+  *(unsigned int *)  (function + 8) = 0x28C0818E;
+  *(unsigned int *)  (function +12) = 0x29C001AE;
+  *(unsigned int *)  (function +16) = 0x28C0A18C;
+  *(unsigned int *)  (function +20) = 0x4C000180;
+  *(unsigned long *) (function +24) = (unsigned long) variable;
+  *(unsigned long *) (function +32) = (unsigned long) data;
+  *(unsigned long *) (function +40) = (unsigned long) address;
+#define TRAMP_CODE_LENGTH  24
+#define is_tramp(function)  \
+  *(unsigned int *)  (function + 0) == 0x1C00000C && \
+  *(unsigned int *)  (function + 4) == 0x28C0618D && \
+  *(unsigned int *)  (function + 8) == 0x28C0818E && \
+  *(unsigned int *)  (function +12) == 0x29C001AE && \
+  *(unsigned int *)  (function +16) == 0x28C0A18C && \
+  *(unsigned int *)  (function +20) == 0x4C000180
+#define tramp_address(function)  \
+  *(unsigned long *) (function +40)
+#define tramp_variable(function)  \
+  *(unsigned long *) (function +24)
+#define tramp_data(function)  \
+  *(unsigned long *) (function +32)
+#endif
 
   /* 3. Set memory protection to "executable" */
 
@@ -1476,11 +1565,7 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
     start_addr = start_addr & -pagesize;
     end_addr = (end_addr + pagesize-1) & -pagesize;
    {uintptr_t len = end_addr - start_addr;
-#if defined(HAVE_MACH_VM)
-    if (vm_protect(task_self(),start_addr,len,0,VM_PROT_READ|VM_PROT_WRITE|VM_PROT_EXECUTE) != KERN_SUCCESS)
-#else
     if (mprotect((void*)start_addr, len, PROT_READ|PROT_WRITE|PROT_EXEC) < 0)
-#endif
       { fprintf(stderr,"trampoline: cannot make memory executable\n"); abort(); }
   }}
 #endif
@@ -1497,6 +1582,29 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
 #endif
 #if !(defined(__hppanew__) || defined(__hppa64new__) || defined(__powerpcaix__) || defined(__powerpc64aix__) || defined(__ia64__))
   /* Only needed if we really set up machine instructions. */
+  /* Use the operating system provided function, when available. */
+#if defined _WIN32 && !defined __CYGWIN__
+  /* Native Windows.
+     FlushInstructionCache
+     <https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-flushinstructioncache>  */
+  HANDLE process = GetCurrentProcess ();
+  while (!FlushInstructionCache (process, function_x, TRAMP_CODE_LENGTH))
+    ;
+#elif defined __APPLE__ && defined __MACH__
+  /* macOS  */
+  sys_icache_invalidate (function_x, TRAMP_CODE_LENGTH);
+#elif defined _AIX
+  /* AIX.  */
+  _sync_cache_range (function_x, TRAMP_CODE_LENGTH);
+#elif defined __sgi
+  /* IRIX.  */
+  cacheflush (function_x, TRAMP_CODE_LENGTH, ICACHE);
+#elif defined __sun
+  /* Solaris.  */
+  extern void sync_instruction_memory (char *, size_t);
+  sync_instruction_memory (function_x, TRAMP_CODE_LENGTH);
+#else
+  /* No operating system provided function. Dispatch according to the CPU. */
 #ifdef __i386__
 #if defined(_WIN32)
   while (!FlushInstructionCache(GetCurrentProcess(),function_x,TRAMP_CODE_LENGTH))
@@ -1581,6 +1689,19 @@ trampoline_function_t alloc_trampoline (trampoline_function_t address, void** va
 #if defined(__powerpc__) || defined(__powerpc64__)
   __TR_clear_cache(function_x);
 #endif
+#if defined(__riscv32__) || defined(__riscv64__)
+#if defined(__linux__)
+  /* Use the libc function. */
+  __riscv_flush_icache((void*)function_x,(void*)(function_x+TRAMP_CODE_LENGTH),0);
+#elif defined(__GNUC__)
+  __asm__ __volatile__ ("fence.i");
+#endif
+#endif
+#if defined(__loongarch64__)
+  /* Use the GCC built-in. It expands to 'ibar 0'. */
+  __clear_cache((void*)function_x,(void*)(function_x+TRAMP_CODE_LENGTH));
+#endif
+#endif
 #endif
 
   /* 5. Return. */
@@ -1599,7 +1720,9 @@ void free_trampoline (trampoline_function_t function)
     function -= ((intptr_t*)page_x)[0];
   }
 #endif
+  gl_lock_lock(freelist_lock);
   *(char**)function = freelist; freelist = (char*)function;
+  gl_lock_unlock(freelist_lock);
   /* It is probably not worth calling munmap() for entirely freed pages. */
 #else
   free(((char**)function)[-1]);
