@@ -1,7 +1,7 @@
 /* vacall function for arm CPU with -mfloat-abi=hard */
 
 /*
- * Copyright 1995-2017 Bruno Haible <bruno@clisp.org>
+ * Copyright 1995-2021 Bruno Haible <bruno@clisp.org>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -50,8 +50,8 @@ register double darg4   __asm__("d4"); /* overlaps s8,s9 */
 register double darg5   __asm__("d5"); /* overlaps s10,s11 */
 register double darg6   __asm__("d6"); /* overlaps s12,s13 */
 register double darg7   __asm__("d7"); /* overlaps s14,s15 */
-register __vaword	iret	__asm__("r0");
-register __vaword	iret2	__asm__("r1");
+register __varword	iret	__asm__("r0");
+register __varword	iret2	__asm__("r1");
 register float		fret	__asm__("s0");
 register double		dret	__asm__("d0");
 
@@ -160,8 +160,8 @@ vacall_receiver (__vaword ignored1, __vaword ignored2, __vaword ignored3, __vawo
     iret = list.tmp._ulong;
   } else
   if (list.rtype == __VAlonglong || list.rtype == __VAulonglong) {
-    iret  = ((__vaword *) &list.tmp._longlong)[0];
-    iret2 = ((__vaword *) &list.tmp._longlong)[1];
+    iret  = ((__varword *) &list.tmp._longlong)[0];
+    iret2 = ((__varword *) &list.tmp._longlong)[1];
   } else
   if (list.rtype == __VAfloat) {
     fret = list.tmp._float;
